"""JSON encoding functions using RuleDispatch."""

import datetime

try:
    import decimal
except ImportError: # Python 2.3
    decimal = None

import dispatch
from simplejson import JSONEncoder
from turbojson.prioritized_methods import PriorityDisambiguated

try:
    from turbogears import config
except ImportError:
    class config(object):
        _default = object()
        def get(cls, key, default=_default):
            if default is cls._default:
                raise KeyError(key)
            return default
        get = classmethod(get)


# Global options

descent_bases = True


# Specific encoding functions

def jsonify(obj):
    """Generic function for converting objects to JSON.

    Specific functions should return a string or an object that can be
    serialized with JSON, i.e., it is made up of only lists, dictionaries
    (with string keys), and strings, ints, and floats.

    """
    raise NotImplementedError
jsonify = dispatch.generic(PriorityDisambiguated)(jsonify)


def jsonify_datetime(obj):
    """JSONify datetime and date objects."""
    return str(obj)
jsonify_datetime = jsonify.when(
        "isinstance(obj, (datetime.datetime, datetime.date))", prio=-1)(
    jsonify_datetime)


def jsonify_decimal(obj):
    """JSONify decimal objects."""
    return float(obj)
if decimal:
    jsonify_decimal = jsonify.when(
        "isinstance(obj, decimal.Decimal)", prio=-1)(jsonify_decimal)


def jsonify_explicit(obj):
    """JSONify objects with explicit JSONification method."""
    return obj.__json__()
jsonify_explicit = jsonify.when(
    "hasattr(obj, '__json__')", prio=-1)(jsonify_explicit)


# SQLObject support

try:
    from sqlobject import SQLObject

    def is_sqlobject(obj):
        return (isinstance(obj, SQLObject)
            and hasattr(obj.__class__, 'sqlmeta'))

    def jsonify_sqlobject(obj):
        """JSONify SQLObjects."""
        result = {}
        result['id'] = obj.id
        keys = []
        sm = obj.__class__.sqlmeta
        try:
            while sm is not None:
                # we need to exclude the ID-keys, as for some reason
                # this won't work for subclassed items
                keys.extend([key
                    for key in sm.columns.keys() if key[-2:] != 'ID'])
                if descent_bases:
                    sm = sm.__base__
                else:
                    sm = None
        except AttributeError:
            # this happens if we descent to <type object>
            pass
        for name in keys:
            result[name] = getattr(obj, name)
        return result
    jsonify_sqlobject = jsonify.when(
            "is_sqlobject(obj) and not hasattr(obj, '__json__')", prio=-1)(
        jsonify_sqlobject)

    try:
        SelectResultsClass = SQLObject.SelectResultsClass
    except AttributeError:
        pass
    else:

        def jsonify_select_results(obj):
            """JSONify SQLObject.SelectResults."""
            return list(obj)
        jsonify_select_results = jsonify.when(
                "isinstance(obj, SelectResultsClass)", prio=-1)(
            jsonify_select_results)

except ImportError:
    pass


# SQLAlchemy support

try:
    import sqlalchemy

    try:
        import sqlalchemy.ext.selectresults
        from sqlalchemy.util import OrderedProperties
    except ImportError: # SQLAlchemy >= 0.5

        def is_saobject(obj):
            return hasattr(obj, '_sa_class_manager')

        def jsonify_saobject(obj):
            """JSONify SQLAlchemy objects."""
            props = {}
            for key in obj.__dict__:
                if not key.startswith('_sa_'):
                    props[key] = getattr(obj, key)
            return props
        jsonify_saobject = jsonify.when(
                "is_saobject(obj) and not hasattr(obj, '__json__')", prio=-1)(
            jsonify_saobject)

    else: # SQLAlchemy < 0.5

        def is_saobject(obj):
            return (hasattr(obj, 'c')
                and isinstance(obj.c, OrderedProperties))

        def jsonify_saobject(obj):
            """JSONify SQLAlchemy objects."""
            props = {}
            for key in obj.c.keys():
                props[key] = getattr(obj, key)
            return props
        jsonify_saobject = jsonify.when(
                "is_saobject(obj) and not hasattr(obj, '__json__')", prio=-1)(
            jsonify_saobject)

        try:
            from sqlalchemy.orm.attributes import InstrumentedList
        except ImportError: # SQLAlchemy >= 0.4
            pass # normal lists are used here

        else: # SQLAlchemy < 0.4

            def jsonify_instrumented_list(obj):
                """JSONify SQLAlchemy instrumented lists."""
                return list(obj)
            jsonify_instrumented_list = jsonify.when(
                    "isinstance(obj, InstrumentedList)", prio=-1)(
                jsonify_instrumented_list)

except ImportError:
    pass


# JSON Encoder class

class GenericJSON(JSONEncoder):

    def __init__(self, **opts):
        opt = opts.pop('descent_bases', None)
        if opt is not None:
            global descent_bases
            descent_bases = opt
        super(GenericJSON, self).__init__(**opts)

    def default(self, obj):
        return jsonify(obj)

_instance = GenericJSON()


# General encoding functions

def encode(obj):
    """Return a JSON string representation of a Python object."""
    return _instance.encode(obj)

def encode_iter(obj):
    """Encode object, yielding each string representation as available."""
    return _instance.iterencode(obj)



from unittest import TestCase
import dispatch
from turbojson.prioritized_methods import PriorityDisambiguated


class TestPriorityDisambiguated(TestCase):

    def setUp(self):
        self.demo_func.clear()
        self.counter = 0

    def demo_func(self, number):
        pass
    demo_func = dispatch.generic(PriorityDisambiguated)(demo_func)

    def test_no_ambiguous(self):
        """Non-ambiguous methods are processed fine"""
        self.demo_func.when("(number >= 0) and (number < 10)")(lambda s, x: 0)
        self.demo_func.when("(number >= 10) and (number < 20)")(lambda s, x: 10)
        self.demo_func.when("(number >= 20) and (number < 30)")(lambda s, x: 20)

        self.assertEqual(self.demo_func(3), 0)
        self.assertEqual(self.demo_func(13), 10)
        self.assertEqual(self.demo_func(23), 20)

    def test_non_prioritized_ambiguous(self):
        """Non-prioritized ambiguous methods are processed fine"""
        self.demo_func.when("(number >= 0) and (number <= 10)")(lambda s, x: 0)
        self.demo_func.when("(number >= 10) and (number < 20)")(lambda s, x: 10)
        self.demo_func.when("(number >= 20) and (number < 30)")(lambda s, x: 20)

        self.assertEqual(self.demo_func(3), 0)
        self.assertRaises(dispatch.AmbiguousMethod, self.demo_func, 10)
        self.assertEqual(self.demo_func(13), 10)
        self.assertEqual(self.demo_func(23), 20)

    def test_no_applicable_methods(self):
        """NoApplicableMethods is raised when it should"""
        self.demo_func.when("(number >= 0) and (number < 10)")(lambda s, x: 0)
        self.demo_func.when("(number >= 10) and (number < 20)")(lambda s, x: 10)
        self.demo_func.when("(number >= 20) and (number < 30)")(lambda s, x: 20)

        self.assertRaises(dispatch.NoApplicableMethods, self.demo_func, -1)
        #self.assertRaises(dispatch.NoApplicableMethods, self.demo_func, 33) # Fails with RuleDispatch pre r2240

    def test_prioritized_ambiguous(self):
        """Prioritized ambiguous methods are processed fine"""
        self.demo_func.when("(number >= 0) and (number <= 10)",prio=1)(
            lambda s,x: 0
        )
        self.demo_func.when("(number >= 10) and (number < 20)")(lambda s, x: 10)
        self.demo_func.when("(number >= 20) and (number < 30)")(lambda s, x: 20)

        self.assertEqual(self.demo_func(3), 0)
        self.assertEqual(self.demo_func(10), 0)
        self.assertEqual(self.demo_func(13), 10)
        self.assertEqual(self.demo_func(23), 20)
        self.assertRaises(dispatch.NoApplicableMethods, self.demo_func, -1)
        #self.assertRaises(dispatch.NoApplicableMethods, self.demo_func, 33) # Fails with RuleDispatch pre r2240

    def test_before(self):
        """Priority is respected when decorating with before"""

        def between_0_30(self, number):
            return number
        between_0_30 = self.demo_func.when(
            "(number >= 0) and (number < 30)")(between_0_30)

        def before_between_2_4(self, number):
            self.counter = number
        before_between_2_4 = self.demo_func.before(
            "(number >= 2) and (number < 4)")(before_between_2_4)

        def before_between_0_30_1(self, number):
            self.counter += number
        before_between_0_30_1 = self.demo_func.before(
            "(number >= 0) and (number < 30)", prio=1)(before_between_0_30_1)

        def before_between_0_30_2(self, number):
            self.counter *= number
        before_between_0_30_2 = self.demo_func.before(
            "(number >= 0) and (number < 30)")(before_between_0_30_2)

        self.assertEqual(self.demo_func(3), 3)
        self.assertEqual(self.counter, 18)



""""prioritized_methods

This is an extension to RuleDispatch to prioritize methods in order
to avoid AmbiguousMethods situations.

Copied from ToscaWidgets, thanks to Alberto Valverde Gonzalez.

"""

import sys

from dispatch import strategy, functions
from dispatch.interfaces import ICriterion, IDispatchPredicate, AmbiguousMethod
from peak.util.decorators import decorate_assignment, frameinfo, decorate_class

__all__ = ['PriorityDisambiguated', 'default_rule', 'generic']

try:
    set
except NameError: # Python 2.3
    from sets import Set as set

default_rule = strategy.default


def _prioritized_safe_methods(grouped_cases):
    """Yield methods in 'grouped_cases', strict version.

    Yield all methods in 'grouped_cases' including those in groups with
    more than one case (AmbiguousMethods) in the priority specified by the
    priority defined when decorating the method.

    Priorities must be distinct within a group of ambiguous methods or else
    an AmbiguousMethod will be raised.

    """
    for group in grouped_cases:
        if len(group) > len(set([meth._prio for sig, meth in group])):
            yield AmbiguousMethod(group)
            break
        elif len(group) > 1:
            methods = [(-meth._prio, meth) for sig, meth in group]
            methods.sort()
            for prio, meth in methods:
                yield meth
        else:
            yield group[0][1]


def _prioritized_all_methods(grouped_cases):
    """Yield methods in 'grouped_cases', loose version.

    Yield all methods in 'grouped_cases' including those in groups with
    more than one case (AmbiguousMethods) in the priority specified by the
    priority defined when decorating the method.

    """
    for group in grouped_cases:
        methods = [(-meth._prio, meth) for sig, meth in group]
        methods.sort()
        for prio, meth in methods:
            yield meth


class PriorityDisambiguated(functions.GenericFunction):
    """GenericFunction with additional priority parameter.

    Remimplementation of dispatch.functions.GenericFunction allowing the case
    decorators to specify a priority for ordering and disambiguation purposes.

    """

    def combine(self, cases):
        """Combine method using priority-ordering generators.

        Same as GenericFunction.combine but replaces strategy.safe_methods
        and strategy.all_methods with priority-ordering generators.

        """
        strict = [strategy.ordered_signatures, _prioritized_safe_methods]
        loose  = [strategy.ordered_signatures, _prioritized_all_methods]
        cases = strategy.separate_qualifiers(cases,
            around=strict, before=loose, primary=strict, after=loose)
        primary = strategy.method_chain(cases.get('primary', []))
        if cases.get('after') or cases.get('before'):
            befores = strategy.method_list(cases.get('before', []))
            afters = strategy.method_list(list(cases.get('after', []))[::-1])
            def chain(*args, **kw):
                for tmp in befores(*args, **kw):
                    pass # toss return values
                result = primary(*args, **kw)
                for tmp in afters(*args, **kw):
                    pass # toss return values
                return result
        else:
            chain = primary
        if cases.get('around'):
            chain = strategy.method_chain(list(cases['around']) + [chain])
        return chain

    def around(self, cond, prio=0):
        """Add function as an 'around' method with 'cond' as a guard.

        If 'cond' is parseable, it will be parsed using the caller's frame
        locals and globals.

        """
        return self._decorate(cond, 'around', prio=prio)

    def before(self, cond, prio=0):
        """Add function as a 'before' method with 'cond' as a guard.

        If 'cond' is parseable, it will be parsed using the caller's frame
        locals and globals.

        """
        return self._decorate(cond, 'before', prio=prio)


    def after(self, cond, prio=0):
        """Add function as an 'after' method with 'cond' as a guard.

        If 'cond' is parseable, it will be parsed using the caller's frame
        locals and globals.

        """
        return self._decorate(cond, 'after', prio=prio)


    def when(self, cond, prio=0):
        """Add following function to this generic one, with 'cond' as a guard.

        If 'cond' is parseable, it will be parsed using the caller's frame
        locals and globals.

        """
        return self._decorate(cond, 'primary', prio=prio)

    def _decorate(self, cond, qualifier=None, frame=None, depth=2, prio=0):
        """Decorate method with priority parameter.

        Analogous to AbstractGeneric._decorate but accepts a priority
        parameter to attach as an attribute to the decorated method to aid
        in disambiguation/ordering.

        """
        frame = frame or sys._getframe(depth)
        rule = cond
        cond = self.parseRule(cond, frame=frame) or cond

        def registerMethod(frm, name, value, old_locals):
            kind, module, locals_, globals_ = frameinfo(frm)
            if qualifier is None:
                func = value
            else:
                func = qualifier, value
            if kind == 'class':
                # 'when()' in class body; defer adding the method
                def registerClassSpecificMethod(cls):
                    req = strategy.Signature(
                        [(strategy.Argument(0), ICriterion(cls))])
                    self.addMethod(req & cond, func, prio=prio)
                    return cls
                decorate_class(registerClassSpecificMethod, frame=frm)
            else:
                self.addMethod(cond, func, prio=prio)
            if old_locals.get(name) in (self, self.delegate):
                return self.delegate
            return value

        return decorate_assignment(registerMethod, frame=frame)

    def addMethod(self, predicate, function, qualifier=None, prio=0):
        if isinstance(function, tuple):
            function[1]._prio = prio
        else:
            function._prio = prio
        if qualifier is not None:
            function = qualifier, function
        for signature in IDispatchPredicate(predicate):
            self[signature] = function
